import numpy
from PyQt5.QtCore import QItemSelection, pyqtSlot
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QKeySequence, QDropEvent, QIcon
from PyQt5.QtWidgets import QHeaderView, QAction, QActionGroup

from urh.models.ProtocolTableModel import ProtocolTableModel
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.Participant import Participant
from urh.ui.views.TableView import TableView


class ProtocolTableView(TableView):
    show_interpretation_clicked = pyqtSignal(int, int, int, int)
    selection_changed = pyqtSignal()
    protocol_view_change_clicked = pyqtSignal(int)
    row_visibility_changed = pyqtSignal()
    writeable_changed = pyqtSignal(bool)
    crop_sync_clicked = pyqtSignal()
    revert_sync_cropping_wanted = pyqtSignal()
    files_dropped = pyqtSignal(list)
    participant_changed = pyqtSignal()
    new_messagetype_clicked = pyqtSignal(list)  # list of protocol messages
    messagetype_selected = pyqtSignal(MessageType, list)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.Fixed)

        self.ref_message_action = QAction(self.tr("Mark as reference message"), self)
        self.ref_message_action.setShortcut(QKeySequence("R"))
        self.ref_message_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.ref_message_action.triggered.connect(self.set_ref_message)

        self.hide_row_action = QAction("Hide selected rows", self)
        self.hide_row_action.setShortcut(QKeySequence("H"))
        self.hide_row_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.hide_row_action.triggered.connect(self.hide_rows)

        self.addAction(self.ref_message_action)
        self.addAction(self.hide_row_action)

        self.zero_hide_offsets = dict()

    def model(self) -> ProtocolTableModel:
        return super().model()

    @property
    def selected_messages(self):
        messages = self.model().protocol.messages
        rows = set(i.row() for i in self.selectionModel().selectedIndexes())
        return [messages[i] for i in rows]

    def selectionChanged(
        self, selection_1: QItemSelection, selection_2: QItemSelection
    ):
        self.selection_changed.emit()
        super().selectionChanged(selection_1, selection_2)

    def dragMoveEvent(self, event):
        event.accept()

    def dragEnterEvent(self, event):
        event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent):
        if len(event.mimeData().urls()) > 0:
            self.files_dropped.emit(event.mimeData().urls())

    def create_context_menu(self):
        menu = super().create_context_menu()
        row = self.rowAt(self.context_menu_pos.y())
        cols = [
            index.column()
            for index in self.selectionModel().selectedIndexes()
            if index.row() == row
        ]
        cols.sort()

        pos = self.context_menu_pos
        row = self.rowAt(pos.y())
        selected_messages = self.selected_messages
        self.participant_actions = {}

        if len(selected_messages) == 0:
            selected_participant = -1
            selected_message_type = -1
        else:
            selected_participant = selected_messages[0].participant
            selected_message_type = selected_messages[0].message_type
            for message in selected_messages:
                if selected_participant != message.participant:
                    selected_participant = -1
                if selected_message_type != message.message_type:
                    selected_message_type = -1
                if selected_message_type == -1 and selected_participant == -1:
                    break

        message_type_menu_str = self.tr("Message type")
        if selected_message_type != -1:
            message_type_menu_str += self.tr(" (" + selected_message_type.name + ")")
        message_type_menu = menu.addMenu(message_type_menu_str)
        message_type_menu.setIcon(QIcon(":/icons/icons/message_type.svg"))
        message_type_group = QActionGroup(self)
        self.message_type_actions = {}

        for message_type in self.model().protocol.message_types:
            action = message_type_menu.addAction(message_type.name)
            action.setCheckable(True)
            action.setActionGroup(message_type_group)

            if selected_message_type == message_type:
                action.setChecked(True)

            self.message_type_actions[action] = message_type
            action.triggered.connect(self.on_message_type_action_triggered)

        new_message_type_action = message_type_menu.addAction("Create new")
        new_message_type_action.setIcon(QIcon.fromTheme("list-add"))
        new_message_type_action.triggered.connect(
            self.on_new_message_type_action_triggered
        )

        if (
            self.model().participants
            and self.model().protocol
            and not self.selection_is_empty
        ):
            participant_group = QActionGroup(self)
            participant_menu_str = self.tr("Participant")
            if selected_participant is None:
                participant_menu_str += self.tr(" (None)")
            elif isinstance(selected_participant, Participant):
                # Ensure we have correct type as selected_participant can be -1 if multiple participants are selected
                participant_menu_str += " (" + selected_participant.name + ")"

            participant_menu = menu.addMenu(participant_menu_str)
            none_participant_action = participant_menu.addAction("None")
            none_participant_action.setCheckable(True)
            none_participant_action.setActionGroup(participant_group)
            none_participant_action.triggered.connect(
                self.on_none_participant_action_triggered
            )

            if selected_participant is None:
                none_participant_action.setChecked(True)

            for participant in self.model().participants:
                pa = participant_menu.addAction(
                    participant.name + " (" + participant.shortname + ")"
                )
                pa.setCheckable(True)
                pa.setActionGroup(participant_group)
                if selected_participant == participant:
                    pa.setChecked(True)

                self.participant_actions[pa] = participant
                pa.triggered.connect(self.on_participant_action_triggered)

        menu.addSeparator()

        if not self.selection_is_empty:
            menu.addAction(self.copy_action)

        menu.addAction(self.hide_row_action)
        hidden_rows = self.model().hidden_rows
        if len(hidden_rows) > 0:
            show_row_action = menu.addAction(
                self.tr("Show all rows (reset {0:d} hidden)".format(len(hidden_rows)))
            )
            show_row_action.triggered.connect(self.on_show_row_action_triggered)

        if self.model().refindex != -1:
            menu.addAction(self.ref_message_action)

        if not self.model().is_writeable:
            show_interpretation_action = menu.addAction(
                self.tr("Show selection in Interpretation")
            )
            show_interpretation_action.setIcon(QIcon.fromTheme("zoom-select"))
            show_interpretation_action.triggered.connect(
                self.on_show_in_interpretation_action_triggered
            )

        if self.model().is_writeable:
            writeable_action = menu.addAction(self.tr("Writeable"))
            writeable_action.setCheckable(True)
            writeable_action.setChecked(True)
        else:
            writeable_action = menu.addAction(
                self.tr("Writeable (decouples from signal)")
            )
            writeable_action.setCheckable(True)
            writeable_action.setChecked(False)

        writeable_action.triggered.connect(self.on_writeable_action_triggered)

        menu.addSeparator()
        undo_stack = self.model().undo_stack
        view = self.model().proto_view

        for plugin in self.controller.plugin_manager.protocol_plugins:
            if plugin.enabled:
                act = plugin.get_action(
                    self,
                    undo_stack,
                    self.selection_range(),
                    self.controller.proto_analyzer,
                    view,
                )
                if act is not None:
                    menu.addAction(act)

                if hasattr(plugin, "zero_hide_offsets"):
                    self.zero_hide_offsets = plugin.command.zero_hide_offsets

        return menu

    @pyqtSlot()
    def set_ref_message(self):
        if self.model().refindex == -1:
            return

        if self.context_menu_pos is None:
            max_row = numpy.max([index.row() for index in self.selectedIndexes()])
            self.model().refindex = max_row
        else:
            self.model().refindex = self.rowAt(self.context_menu_pos.y())

    def set_row_visibility_status(self, show: bool, rows=None):
        if rows is None:
            rows = self.selected_rows
        elif isinstance(rows, set) or isinstance(rows, list) or isinstance(rows, range):
            rows = rows
        else:
            rows = [rows]

        refindex = self.model().refindex
        for row in rows:
            if show:
                self.showRow(row)
                self.model().hidden_rows.discard(row)
            else:
                if row == refindex:
                    refindex += 1
                self.hideRow(row)
                self.model().hidden_rows.add(row)

        self.model().refindex = refindex
        self.model().update()
        self.row_visibility_changed.emit()

    def show_rows(self, rows=None):
        self.set_row_visibility_status(show=True, rows=rows)

    @pyqtSlot()
    def hide_rows(self, row=None):
        self.set_row_visibility_status(show=False, rows=row)

    @pyqtSlot()
    def on_bit_action_triggered(self):
        self.protocol_view_change_clicked.emit(0)

    @pyqtSlot()
    def on_hex_action_triggered(self):
        self.protocol_view_change_clicked.emit(1)

    @pyqtSlot()
    def on_ascii_action_triggered(self):
        self.protocol_view_change_clicked.emit(2)

    @pyqtSlot()
    def on_none_participant_action_triggered(self):
        for message in self.selected_messages:
            message.participant = None
        self.participant_changed.emit()

    @pyqtSlot()
    def on_participant_action_triggered(self):
        for message in self.selected_messages:
            message.participant = self.participant_actions[self.sender()]
        self.participant_changed.emit()

    @pyqtSlot()
    def on_message_type_action_triggered(self):
        self.messagetype_selected.emit(
            self.message_type_actions[self.sender()], self.selected_messages
        )

    @pyqtSlot()
    def on_new_message_type_action_triggered(self):
        self.new_messagetype_clicked.emit(self.selected_messages)

    @pyqtSlot()
    def on_show_in_interpretation_action_triggered(self):
        min_row, max_row, start, end = self.selection_range()

        offsets = self.zero_hide_offsets.get(min_row, dict())
        start += sum(offsets[i] for i in offsets if i <= start)
        end += sum(offsets[i] for i in offsets if i <= end)

        self.show_interpretation_clicked.emit(min_row, start, max_row, end - 1)

    @pyqtSlot()
    def on_show_row_action_triggered(self):
        for i in self.model().hidden_rows:
            self.showRow(i)
        self.model().hidden_rows.clear()
        self.model().update()
        self.row_visibility_changed.emit()

    @pyqtSlot()
    def on_writeable_action_triggered(self):
        self.writeable_changed.emit(self.sender().isChecked())



import os
import shutil
import sys
import tempfile
from collections import defaultdict
from distutils import ccompiler
from importlib import import_module
from subprocess import check_output

from setuptools import Extension

USE_RELATIVE_PATHS = False

COMPILER_DIRECTIVES = {
    "language_level": 3,
    "cdivision": True,
    "wraparound": False,
    "boundscheck": False,
    "initializedcheck": False,
}

DEVICES = {
    "airspy": {"lib": "airspy", "test_function": "open"},
    "bladerf": {
        "lib": "bladeRF",
        "test_function": "bladerf_open",
        "api_header_include": "libbladeRF.h",
        "api_version_check_code": """
                    #include<stdio.h>
                    #include<libbladeRF.h>
                    
                    int main(void) {
                    struct bladerf_version result; bladerf_version(&result);
                    printf("%f", result.major + result.minor/10.0 + result.patch/100.0);
                    return 0;}
                    """,
    },
    "hackrf": {
        "lib": "hackrf",
        "test_function": "hackrf_init",
        "extras": {"HACKRF_MULTI_DEVICE_SUPPORT": "hackrf_open_by_serial"},
    },
    "limesdr": {"lib": "LimeSuite", "test_function": "LMS_GetDeviceList"},
    "plutosdr": {"lib": "iio", "test_function": "iio_create_default_context"},
    "rtlsdr": {
        "lib": "rtlsdr",
        "test_function": "rtlsdr_get_device_name",
        "extras": {"RTLSDR_BANDWIDTH_SUPPORT": "rtlsdr_set_tuner_bandwidth"},
    },
    # Use C only for USRP to avoid boost dependency
    "usrp": {"lib": "uhd", "test_function": "uhd_usrp_find", "language": "c"},
    "sdrplay": {
        "lib": "mir_sdr_api" if sys.platform == "win32" else "mirsdrapi-rsp",
        "test_function": "mir_sdr_ApiVersion",
    },
}


def compiler_has_function(
    compiler, function_name, libraries, library_dirs, include_dirs
) -> bool:
    tmp_dir = tempfile.mkdtemp(prefix="urh-")
    devnull = old_stderr = None
    try:
        try:
            file_name = os.path.join(tmp_dir, "{}.c".format(function_name))
            with open(file_name, "w") as f:
                # declare function in order to prevent Clang 12 error (https://github.com/jopohl/urh/issues/811)
                f.write("void %s();\n" % function_name)
                f.write("int main(void) {\n")
                f.write("    %s();\n" % function_name)
                f.write("}\n")

            # Redirect stderr to /dev/null to hide any error messages from the compiler.
            devnull = open(os.devnull, "w")
            old_stderr = os.dup(sys.stderr.fileno())
            os.dup2(devnull.fileno(), sys.stderr.fileno())
            objects = compiler.compile([file_name], include_dirs=include_dirs)
            compiler.link_executable(
                objects,
                os.path.join(tmp_dir, "a.out"),
                library_dirs=library_dirs,
                libraries=libraries,
            )
        except Exception as e:
            return False
        return True
    finally:
        if old_stderr is not None:
            os.dup2(old_stderr, sys.stderr.fileno())
        if devnull is not None:
            devnull.close()
        shutil.rmtree(tmp_dir)


def check_api_version(
    compiler, api_version_code, libraries, library_dirs, include_dirs
) -> float:
    tmp_dir = tempfile.mkdtemp(prefix="urh-")
    try:
        try:
            file_name = os.path.join(tmp_dir, "get_api_version.c")
            with open(file_name, "w") as f:
                f.write(api_version_code)

            objects = compiler.compile([file_name], include_dirs=include_dirs)
            check_api_program = os.path.join(tmp_dir, "check_api")
            compiler.link_executable(
                objects,
                check_api_program,
                library_dirs=library_dirs,
                libraries=libraries,
            )

            env = os.environ.copy()
            env["PATH"] = (
                os.pathsep.join(library_dirs) + os.pathsep + os.environ.get("PATH", "")
            )

            result = float(check_output(check_api_program, env=env))
            print("    Automatic API version check succeeded.")
            return result
        except Exception as e:
            print("    API version check failed: {}".format(e))
            return 0.0
    finally:
        shutil.rmtree(tmp_dir)


def get_device_extensions_and_extras(library_dirs=None, include_dirs=None):
    library_dirs = [] if library_dirs is None else library_dirs

    cur_dir = os.path.dirname(os.path.realpath(__file__))
    include_dirs = [] if include_dirs is None else include_dirs

    device_extras = dict()

    if os.path.isdir(os.path.join(cur_dir, "lib/shared")):
        # Device libs are packaged, so we are in release mode
        include_dirs.insert(
            0, os.path.realpath(os.path.join(cur_dir, "lib/shared/include"))
        )
        library_dirs.insert(0, os.path.realpath(os.path.join(cur_dir, "lib/shared")))

    if sys.platform == "darwin":
        for prefix in ["/usr/local", "/opt/homebrew"]:
            include_dirs.append(prefix + "/include")
            library_dirs.append(prefix + "/lib")

    result = []

    # None = automatic (depending on lib is installed)
    # 1 = install extension always
    # 0 = Do not install extension
    build_device_extensions = defaultdict(lambda: None)

    for dev_name in DEVICES:
        with_option = "--with-" + dev_name
        without_option = "--without-" + dev_name

        if with_option in sys.argv and without_option in sys.argv:
            print("ambiguous options for " + dev_name)
            sys.exit(1)
        elif without_option in sys.argv:
            build_device_extensions[dev_name] = 0
            sys.argv.remove(without_option)
        elif with_option in sys.argv:
            build_device_extensions[dev_name] = 1
            sys.argv.remove(with_option)

    sys.path.append(os.path.realpath(os.path.join(cur_dir, "lib")))

    compiler = ccompiler.new_compiler()
    for dev_name, params in DEVICES.items():
        if build_device_extensions[dev_name] == 0:
            print("Skipping native {0} support".format(dev_name))
            continue

        if build_device_extensions[dev_name] == 1:
            print("Enforcing native {0} support".format(dev_name))
        elif compiler_has_function(
            compiler,
            params["test_function"],
            (params["lib"],),
            library_dirs,
            include_dirs,
        ):
            print(
                "Found {0} lib. Will compile with native {1} support".format(
                    params["lib"], dev_name
                )
            )
        else:
            print("Skipping native support for {0}".format(dev_name))
            continue

        device_extras.update(
            get_device_extras(
                compiler, dev_name, [params["lib"]], library_dirs, include_dirs
            )
        )
        if "api_version_check_code" in params:
            env_name = dev_name.upper() + "_API_VERSION"
            ver = os.getenv(env_name)
            if ver is not None:
                try:
                    ver = float(ver)
                except Exception as e:
                    print(
                        "    Could not convert content of {} to float: {}".format(
                            env_name, e
                        )
                    )
                    print("    Will now try to automatically detect API version.")
                    ver = None
            else:
                print(
                    "    Environment variable {} is unset, try to automatically detect API version".format(
                        env_name
                    )
                )

            if ver is None:
                ver = check_api_version(
                    compiler,
                    params["api_version_check_code"],
                    (params["lib"],),
                    library_dirs,
                    include_dirs,
                )
            device_extras[env_name] = ver
            print("    Using {}={}".format(env_name, ver))

        extension = get_device_extension(
            dev_name, [params["lib"]], library_dirs, include_dirs
        )
        result.append(extension)

    return result, device_extras


def get_device_extras(compiler, dev_name, libraries, library_dirs, include_dirs):
    try:
        extras = DEVICES[dev_name]["extras"]
    except KeyError:
        extras = dict()

    result = dict()

    for extra, func_name in extras.items():
        if compiler_has_function(
            compiler, func_name, libraries, library_dirs, include_dirs
        ):
            result[extra] = 1
        else:
            print("Skipping {} as installed driver does not support it".format(extra))
            result[extra] = 0

    return result


def get_device_extension(
    dev_name: str, libraries: list, library_dirs: list, include_dirs: list
):
    try:
        language = DEVICES[dev_name]["language"]
    except KeyError:
        language = "c++"

    cur_dir = os.path.dirname(os.path.realpath(__file__))
    if USE_RELATIVE_PATHS:
        # We need relative paths on windows
        cpp_file_path = "src/urh/dev/native/lib/{0}.pyx".format(dev_name)
    else:
        cpp_file_path = os.path.join(cur_dir, "lib", "{0}.pyx".format(dev_name))

    return Extension(
        "urh.dev.native.lib." + dev_name,
        [cpp_file_path],
        libraries=libraries,
        library_dirs=library_dirs,
        include_dirs=include_dirs,
        language=language,
    )


def perform_health_check() -> str:
    result = []
    for device in sorted(DEVICES.keys()):
        try:
            _ = import_module("urh.dev.native.lib." + device)
            result.append(device + " -- OK")
        except ImportError as e:
            result.append(device + " -- ERROR: " + str(e))

    return "\n".join(result)


if __name__ == "__main__":
    from setuptools import setup

    if "-L" in sys.argv:
        library_directories = sys.argv[sys.argv.index("-L") + 1].split(":")
    else:
        library_directories = None

    if "-I" in sys.argv:
        include_directories = sys.argv[sys.argv.index("-I") + 1].split(":")
    else:
        include_directories = []

    import numpy as np

    include_directories.append(np.get_include())

    cur_dir = os.path.dirname(os.path.realpath(__file__))
    os.chdir(os.path.join(cur_dir, "..", "..", ".."))

    try:
        from Cython.Build import cythonize
    except ImportError:
        print(
            "You need Cython to rebuild URH's device extensions. "
            "You can get it e.g. with python3 -m pip install cython.",
            file=sys.stderr,
        )
        sys.exit(1)

    dev_extensions, dev_extras = get_device_extensions_and_extras(
        library_dirs=library_directories, include_dirs=include_directories
    )
    setup(
        name="urh",
        ext_modules=cythonize(
            dev_extensions,
            force=True,
            compile_time_env=dev_extras,
            compiler_directives=COMPILER_DIRECTIVES,
        ),
    )



import xml.etree.ElementTree as ET

from urh.signalprocessing.Message import Message
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.Participant import Participant
from urh.simulator.SimulatorItem import SimulatorItem
from urh.simulator.SimulatorRule import SimulatorRuleCondition
from urh.util.Formatter import Formatter


class SimulatorMessage(Message, SimulatorItem):
    def __init__(
        self,
        destination: Participant,
        plain_bits,
        pause: int,
        message_type: MessageType,
        decoder=None,
        source=None,
        timestamp=None,
    ):
        Message.__init__(
            self, plain_bits, pause, message_type, decoder=decoder, participant=source
        )
        SimulatorItem.__init__(self)
        if timestamp is not None:
            self.timestamp = timestamp

        self.destination = destination
        self.send_recv_messages = []
        self.repeat = 1

    def set_parent(self, value):
        if value is not None:
            assert value.parent() is None or isinstance(value, SimulatorRuleCondition)

        super().set_parent(value)

    @property
    def source(self):
        return self.participant

    @source.setter
    def source(self, participant):
        self.participant = participant

    @property
    def children(self):
        return self.message_type

    def insert_child(self, pos, child):
        self.children.append(child)
        child.set_parent(self)

    def validate(self):
        return all(child.is_valid for child in self.children)

    @property
    def plain_ascii_str(self) -> str:
        if len(self.send_recv_messages) > 0:
            plain_ascii_array = self.send_recv_messages[-1].plain_ascii_array
        else:
            plain_ascii_array = self.plain_ascii_array

        return "".join(map(chr, plain_ascii_array))

    @property
    def plain_bits_str(self) -> str:
        return (
            str(self.send_recv_messages[-1])
            if len(self.send_recv_messages) > 0
            else str(self)
        )

    def __delitem__(self, index):
        removed_labels = self._remove_labels_for_range(index, instant_remove=False)
        self.simulator_config.delete_items(removed_labels)
        del self.plain_bits[index]

    def to_xml(
        self, decoders=None, include_message_type=False, write_bits=True
    ) -> ET.Element:
        result = ET.Element(
            "simulator_message",
            attrib={
                "destination_id": self.destination.id if self.destination else "",
                "repeat": str(self.repeat),
            },
        )

        result.append(
            super().to_xml(decoders, include_message_type, write_bits=write_bits)
        )

        return result

    def from_xml(
        self, tag: ET.Element, participants, decoders=None, message_types=None
    ):
        super().from_xml(tag, participants, decoders, message_types)
        self.destination = Participant.find_matching(
            tag.get("destination_id", ""), participants
        )
        self.repeat = Formatter.str2val(tag.get("repeat", "1"), int, 1)

    @classmethod
    def new_from_xml(
        cls, tag: ET.Element, participants, decoders=None, message_types=None
    ):
        msg = Message.new_from_xml(
            tag.find("message"),
            participants=participants,
            decoders=decoders,
            message_types=message_types,
        )
        destination = Participant.find_matching(
            tag.get("destination_id", ""), participants
        )
        return SimulatorMessage(
            destination,
            msg.plain_bits,
            msg.pause,
            msg.message_type,
            msg.decoder,
            msg.participant,
            timestamp=msg.timestamp,
        )



cdef extern from "rtl-sdr.h":
    ctypedef struct rtlsdr_dev_t:
        pass

    ctypedef unsigned char uint8_t
    ctypedef unsigned short uint16_t
    ctypedef unsigned int  uint32_t
    ctypedef unsigned long long uint64_t

    uint32_t rtlsdr_get_device_count()
    const char*rtlsdr_get_device_name(uint32_t index)

    int rtlsdr_get_device_usb_strings(uint32_t index, char *manufact, char *product, char *serial)
    int rtlsdr_get_index_by_serial(const char *serial)

    int rtlsdr_open(rtlsdr_dev_t ** dev, uint32_t index)
    int rtlsdr_close(rtlsdr_dev_t *dev)

    int rtlsdr_set_xtal_freq(rtlsdr_dev_t *dev, uint32_t rtl_freq, uint32_t tuner_freq)
    int rtlsdr_get_xtal_freq(rtlsdr_dev_t *dev, uint32_t *rtl_freq, uint32_t *tuner_freq)

    int rtlsdr_get_usb_strings(rtlsdr_dev_t *dev, char *manufact, char *product, char *serial)

    int rtlsdr_write_eeprom(rtlsdr_dev_t *dev, uint8_t *data, uint8_t offset, uint16_t len)
    int rtlsdr_read_eeprom(rtlsdr_dev_t *dev, uint8_t *data, uint8_t offset, uint16_t len)

    int rtlsdr_set_center_freq(rtlsdr_dev_t *dev, uint32_t freq)
    uint32_t rtlsdr_get_center_freq(rtlsdr_dev_t *dev)

    int rtlsdr_set_freq_correction(rtlsdr_dev_t *dev, int ppm);
    int rtlsdr_get_freq_correction(rtlsdr_dev_t *dev);

    IF RTLSDR_BANDWIDTH_SUPPORT == 1:
        int rtlsdr_set_tuner_bandwidth(rtlsdr_dev_t *dev, uint32_t bw);

    ctypedef enum rtlsdr_tuner:
        RTLSDR_TUNER_UNKNOWN = 0,
        RTLSDR_TUNER_E4000,
        RTLSDR_TUNER_FC0012,
        RTLSDR_TUNER_FC0013,
        RTLSDR_TUNER_FC2580,
        RTLSDR_TUNER_R820T,
        RTLSDR_TUNER_R828D

    rtlsdr_tuner rtlsdr_get_tuner_type(rtlsdr_dev_t *dev)

    int rtlsdr_get_tuner_gains(rtlsdr_dev_t *dev, int *gains)
    int rtlsdr_set_tuner_gain(rtlsdr_dev_t *dev, int gain)
    int rtlsdr_get_tuner_gain(rtlsdr_dev_t *dev)
    int rtlsdr_set_tuner_if_gain(rtlsdr_dev_t *dev, int stage, int gain)
    int rtlsdr_set_tuner_gain_mode(rtlsdr_dev_t *dev, int manual)

    int rtlsdr_set_sample_rate(rtlsdr_dev_t *dev, uint32_t rate)
    uint32_t rtlsdr_get_sample_rate(rtlsdr_dev_t *dev)

    int rtlsdr_set_testmode(rtlsdr_dev_t *dev, int on)
    int rtlsdr_set_agc_mode(rtlsdr_dev_t *dev, int on)

    int rtlsdr_set_direct_sampling(rtlsdr_dev_t *dev, int on)
    int rtlsdr_get_direct_sampling(rtlsdr_dev_t *dev)

    int rtlsdr_set_offset_tuning(rtlsdr_dev_t *dev, int on)
    int rtlsdr_get_offset_tuning(rtlsdr_dev_t *dev)

    int rtlsdr_reset_buffer(rtlsdr_dev_t *dev)
    int rtlsdr_read_sync(rtlsdr_dev_t *dev, void *buf, int len, int *n_read)

    ctypedef void (*rtlsdr_read_async_cb_t)(unsigned char *buffer, uint32_t length, void *ctx)

    int rtlsdr_read_async(rtlsdr_dev_t *dev, rtlsdr_read_async_cb_t cb, void *ctx, uint32_t buf_num, uint32_t buf_len);
    int rtlsdr_cancel_async(rtlsdr_dev_t *dev)

    int rtlsdr_set_bias_tee(rtlsdr_dev_t *dev, int on)


from PyQt5.QtWidgets import QAction, QUndoStack, QMessageBox


from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from ..Plugin import ProtocolPlugin
from ..MessageBreak.MessageBreakAction import MessageBreakAction


class MessageBreakPlugin(ProtocolPlugin):
    def __init__(self):
        super().__init__(name="MessageBreak")
        self.undo_stack = None
        self.command = None
        """:type: QUndoAction """

    def get_action(
        self,
        parent,
        undo_stack: QUndoStack,
        sel_range,
        protocol: ProtocolAnalyzer,
        view: int,
    ):
        """
        :type parent: QTableView
        :type undo_stack: QUndoStack
        :type protocol_analyzers: list of ProtocolAnalyzer
        """
        min_row, max_row, start, end = sel_range
        if min_row == -1 or max_row == -1 or start == -1 or end == -1:
            return None

        if max_row != min_row:
            return None

        end = protocol.convert_index(end, view, 0, True, message_indx=min_row)[0]
        # factor = 1 if view == 0 else 4 if view == 1 else 8

        self.command = MessageBreakAction(protocol, max_row, end)
        action = QAction(self.command.text(), parent)
        action.triggered.connect(self.action_triggered)
        self.undo_stack = undo_stack
        return action

    def action_triggered(self):
        self.undo_stack.push(self.command)



# noinspection PyUnresolvedReferences
cimport numpy as np
import cython
import numpy as np
from libcpp cimport bool

from libc.stdint cimport uint8_t, uint16_t, uint32_t, int64_t
from libc.stdio cimport printf
from libc.stdlib cimport malloc, free
from urh.cythonext.util cimport IQ, iq, bit_array_to_number

from cython.parallel import prange
from libc.math cimport atan2, sqrt, M_PI, abs

cdef extern from "math.h" nogil:
    float cosf(float x)
    float acosf(float x)
    float sinf(float x)

cdef extern from "complex.h" namespace "std" nogil:
    float arg(float complex x)
    float complex conj(float complex x)

# As we do not use any numpy C API functions we do no import_array here,
# because it can lead to OS X error: https://github.com/jopohl/urh/issues/273
# np.import_array()

cdef int64_t PAUSE_STATE = -1

cdef float complex imag_unit = 1j
cdef float NOISE_FSK_PSK = -4.0
cdef float NOISE_ASK = 0.0

cdef float get_noise_for_mod_type(str mod_type):
    if mod_type == "ASK":
        return NOISE_ASK
    elif mod_type == "FSK":
        return NOISE_FSK_PSK
    elif mod_type == "PSK" or mod_type == "OQPSK":
        return NOISE_FSK_PSK
    elif mod_type == "QAM":
        return NOISE_ASK * NOISE_FSK_PSK
    else:
        return 0

cdef get_numpy_dtype(iq cython_type):
    if str(cython.typeof(cython_type)) == "char":
        return np.int8
    elif str(cython.typeof(cython_type)) == "short":
        return np.int16
    elif str(cython.typeof(cython_type)) == "float":
        return np.float32
    else:
        raise ValueError("dtype {} not supported for modulation".format(cython.typeof(cython_type)))

cpdef modulate_c(uint8_t[:] bits, uint32_t samples_per_symbol, str modulation_type,
                 float[:] parameters, uint16_t bits_per_symbol,
                 float carrier_amplitude, float carrier_frequency, float carrier_phase, float sample_rate,
                 uint32_t pause, uint32_t start, dtype=np.float32,
                 float gauss_bt=0.5, float filter_width=1.0):

    if dtype == np.int8:
        return __modulate(
            bits, samples_per_symbol, modulation_type, parameters, bits_per_symbol, carrier_amplitude,
            carrier_frequency, carrier_phase, sample_rate, pause, start, <char>0, gauss_bt, filter_width
        )
    elif dtype == np.int16:
        return __modulate(
            bits, samples_per_symbol, modulation_type, parameters, bits_per_symbol, carrier_amplitude,
            carrier_frequency, carrier_phase, sample_rate, pause, start, <short>0, gauss_bt, filter_width
        )
    elif dtype == np.float32:
        return __modulate(
            bits, samples_per_symbol, modulation_type, parameters, bits_per_symbol, carrier_amplitude,
            carrier_frequency, carrier_phase, sample_rate, pause, start, <float>0.0, gauss_bt, filter_width
        )
    else:
        raise ValueError("Unsupported dtype for modulation {}".format(dtype))


cpdef __modulate(uint8_t[:] bits, uint32_t samples_per_symbol, str modulation_type,
                 float[:] parameters, uint16_t bits_per_symbol,
                 float carrier_amplitude, float carrier_frequency, float carrier_phase, float sample_rate,
                 uint32_t pause, uint32_t start, iq iq_type,
                 float gauss_bt=0.5, float filter_width=1.0):
    cdef int64_t i = 0, j = 0, index = 0, prev_index=0, s_i = 0, num_bits = len(bits)
    cdef uint32_t total_symbols = int(num_bits // bits_per_symbol)
    cdef int64_t total_samples = total_symbols * samples_per_symbol + pause

    cdef float a = carrier_amplitude, f = carrier_frequency, phi = carrier_phase

    cdef float f_previous = 0, phase_correction = 0
    cdef float t = 0, current_arg = 0

    result = np.zeros((total_samples, 2), dtype=get_numpy_dtype(iq_type))
    if num_bits == 0:
        return result

    cdef iq[:, ::1] result_view = result

    cdef bool is_fsk = modulation_type.lower() == "fsk"
    cdef bool is_ask = modulation_type.lower() == "ask"
    cdef bool is_psk = modulation_type.lower() == "psk"
    cdef bool is_oqpsk = modulation_type.lower() == "oqpsk"
    cdef bool is_gfsk = modulation_type.lower() == "gfsk"

    assert is_fsk or is_ask or is_psk or is_gfsk or is_oqpsk

    cdef uint8_t[:] oqpsk_bits
    if is_oqpsk:
        assert bits_per_symbol == 2
        bits = get_oqpsk_bits(bits)

    cdef np.ndarray[np.float32_t, ndim=2] gauss_filtered_freqs_phases
    if is_gfsk:
        gauss_filtered_freqs_phases = get_gauss_filtered_freqs_phases(bits, parameters, total_symbols,
                                                                      samples_per_symbol, sample_rate, carrier_phase,
                                                                      start, gauss_bt, filter_width)


    cdef float* phase_corrections = NULL
    if is_fsk and total_symbols > 0:
        phase_corrections = <float*>malloc(total_symbols * sizeof(float))
        phase_corrections[0] = 0.0
        for s_i in range(1, total_symbols):
            # Add phase correction to FSK modulation in order to prevent spiky jumps
            index = bit_array_to_number(bits, end=(s_i+1)*bits_per_symbol, start=s_i*bits_per_symbol)
            prev_index = bit_array_to_number(bits, end=s_i*bits_per_symbol, start=(s_i-1)*bits_per_symbol)

            f = parameters[index]
            f_previous = parameters[prev_index]

            if f != f_previous:
                t = (s_i*samples_per_symbol+start-1) / sample_rate
                phase_corrections[s_i] = (phase_corrections[s_i-1] + 2 * M_PI * (f_previous-f) * t) % (2 * M_PI)
            else:
                phase_corrections[s_i] = phase_corrections[s_i-1]

    for s_i in prange(0, total_symbols, schedule="static", nogil=True):
        index = bit_array_to_number(bits, end=(s_i+1)*bits_per_symbol, start=s_i*bits_per_symbol)

        a = carrier_amplitude
        f = carrier_frequency
        phi = carrier_phase
        phase_correction = 0

        if is_ask:
            a = parameters[index]
            if a == 0:
                continue
        elif is_fsk:
            f = parameters[index]
            phase_correction = phase_corrections[s_i]

        elif is_psk or is_oqpsk:
            phi = parameters[index]

        for i in range(s_i * samples_per_symbol, (s_i+1)*samples_per_symbol):
            t = (i+start) / sample_rate
            if is_gfsk:
                f = gauss_filtered_freqs_phases[i, 0]
                phi = gauss_filtered_freqs_phases[i, 1]
            current_arg = 2 * M_PI * f * t + phi + phase_correction

            result_view[i, 0] = <iq>(a * cosf(current_arg))
            result_view[i, 1] = <iq>(a * sinf(current_arg))

    if is_oqpsk:
        for i in range(0, samples_per_symbol):
            result_view[i, 1] = 0
        for i in range(total_samples-pause-samples_per_symbol, total_samples-pause):
            result_view[i, 0] = 0

    if phase_corrections != NULL:
        free(phase_corrections)

    return result

cpdef uint8_t[:] get_oqpsk_bits(uint8_t[:] original_bits):
    # TODO: This method does not work correctly. Fix it when we have a test signal
    cdef int64_t i, num_bits = len(original_bits)
    if num_bits == 0:
        return np.zeros(0, dtype=np.uint8)

    result = np.zeros(num_bits+2, dtype=np.uint8)
    result[0] = original_bits[0]
    result[num_bits+2-1] = original_bits[num_bits-1]

    for i in range(2, num_bits-2, 2):
        result[i] = original_bits[i]
        result[i+1] = original_bits[i-1]

    return result


cdef np.ndarray[np.float32_t, ndim=2] get_gauss_filtered_freqs_phases(uint8_t[:] bits,  float[:] parameters,
                                                                     uint32_t num_symbols, uint32_t samples_per_symbol,
                                                                     float sample_rate, float phi, uint32_t start,
                                                                     float gauss_bt, float filter_width):
    cdef int64_t i, s_i, index, num_values = num_symbols * samples_per_symbol
    cdef np.ndarray[np.float32_t, ndim=1] frequencies = np.empty(num_values, dtype=np.float32)
    cdef uint16_t bits_per_symbol = int(len(bits) // num_symbols)

    for s_i in range(0, num_symbols):
        index = bit_array_to_number(bits, end=(s_i+1)*bits_per_symbol, start=s_i*bits_per_symbol)

        for i in range(s_i * samples_per_symbol, (s_i+1)*samples_per_symbol):
            frequencies[i] = parameters[index]

    cdef np.ndarray[np.float32_t, ndim=1] t = np.arange(start, start + num_values, dtype=np.float32) / sample_rate
    cdef np.ndarray[np.float32_t, ndim=1] gfir = gauss_fir(sample_rate, samples_per_symbol,
                                                           bt=gauss_bt, filter_width=filter_width)

    if len(frequencies) >= len(gfir):
        frequencies = np.convolve(frequencies, gfir, mode="same")
    else:
        # Prevent dimension crash later, because gaussian finite impulse response is longer then param_vector
        frequencies = np.convolve(gfir, frequencies, mode="same")[:len(frequencies)]

    cdef np.ndarray[np.float32_t, ndim=1] phases = np.zeros(len(frequencies), dtype=np.float32)
    phases[0] = phi
    for i in range(0, len(phases) - 1):
         # Correct the phase to prevent spiky jumps
        phases[i + 1] = 2 * M_PI * t[i] * (frequencies[i] - frequencies[i + 1]) + phases[i]

    return np.column_stack((frequencies, phases))

cdef np.ndarray[np.float32_t, ndim=1] gauss_fir(float sample_rate, uint32_t samples_per_symbol,
                                                float bt=.5, float filter_width=1.0):
    """

    :param filter_width: Filter width
    :param bt: normalized 3-dB bandwidth-symbol time product
    :return:
    """
    # http://onlinelibrary.wiley.com/doi/10.1002/9780470041956.app2/pdf
    cdef np.ndarray[np.float32_t] k = np.arange(-int(filter_width * samples_per_symbol),
                                                int(filter_width * samples_per_symbol) + 1,
                                                dtype=np.float32)
    cdef float ts = samples_per_symbol / sample_rate  # symbol time
    cdef np.ndarray[np.float32_t] h = np.sqrt((2 * np.pi) / (np.log(2))) * bt / ts * np.exp(
        -(((np.sqrt(2) * np.pi) / np.sqrt(np.log(2)) * bt * k / samples_per_symbol) ** 2))
    return h / h.sum()

cdef float clamp(float x) nogil:
    if x < -1.0:
        x = -1.0
    elif x > 1.0:
        x = 1.0
    return x

cdef float[::1] costa_demod(IQ samples, float noise_sqrd, int loop_order, float bandwidth=0.1, float damping=sqrt(2.0) / 2.0):
    cdef float alpha = (4 * damping * bandwidth) / (1.0 + 2.0 * damping * bandwidth + bandwidth * bandwidth)
    cdef float beta = (4 * bandwidth * bandwidth) / (1.0 + 2.0 * damping * bandwidth + bandwidth * bandwidth)

    cdef long long i = 0, num_samples = len(samples)
    cdef float real = 0, imag = 0

    cdef float scale, shift, real_float, imag_float, ref_real, ref_imag

    cdef float f1, f2, costa_freq = 0, costa_error = 0, costa_phase = 1.5

    cdef float complex current_sample, nco_out, nco_times_sample

    cdef float[::1] result = np.empty(num_samples, dtype=np.float32)

    if str(cython.typeof(samples)) == "char[:, ::1]":
        scale = 127.5
        shift = 0.5
    elif str(cython.typeof(samples)) == "unsigned char[:, ::1]":
        scale = 127.5
        shift = -127.5
    elif str(cython.typeof(samples)) == "short[:, ::1]":
        scale = 32767.5
        shift = 0.5
    elif str(cython.typeof(samples)) == "unsigned short[:, ::1]":
        scale = 65535.0
        shift = -32767.5
    elif str(cython.typeof(samples)) == "float[:, ::1]":
        scale = 1.0
        shift = 0.0
    else:
        raise ValueError("Unsupported dtype")

    if loop_order > 4:
        # TODO: Adapt this when PSK demodulation with order > 4 shall be supported
        loop_order = 4

    for i in range(1, num_samples):
        real = samples[i, 0]
        imag = samples[i, 1]

        if real * real + imag * imag <= noise_sqrd:
            result[i] = NOISE_FSK_PSK
            continue

        real_float = (real + shift) / scale
        imag_float = (imag + shift) / scale

        current_sample = real_float + imag_unit * imag_float
        nco_out = cosf(-costa_phase) + imag_unit * sinf(-costa_phase)
        nco_times_sample = nco_out * current_sample

        if loop_order == 2:
            costa_error = nco_times_sample.imag * nco_times_sample.real
        elif loop_order == 4:
            f1 = 1.0 if nco_times_sample.real > 0.0 else -1.0
            f2 = 1.0 if nco_times_sample.imag > 0.0 else -1.0
            costa_error = f1 * nco_times_sample.imag - f2 * nco_times_sample.real

        costa_error = clamp(costa_error)

        # advance the loop
        costa_freq += beta * costa_error
        costa_phase += costa_freq + alpha * costa_error

        # wrap the phase
        while costa_phase > (2 * M_PI):
            costa_phase -= 2 * M_PI
        while costa_phase < (-2 * M_PI):
            costa_phase += 2 * M_PI

        costa_freq = clamp(costa_freq)

        if loop_order == 2:
            result[i] = nco_times_sample.real
        elif loop_order == 4:
            result[i] = 2 * nco_times_sample.real + nco_times_sample.imag

    return result


cpdef np.ndarray[np.float32_t, ndim=1] afp_demod(IQ samples, float noise_mag,
                                                 str mod_type, int mod_order, float costas_loop_bandwidth=0.1):
    if len(samples) <= 2:
        return np.zeros(len(samples), dtype=np.float32)

    cdef long long i = 0, ns = len(samples)
    cdef float NOISE = get_noise_for_mod_type(mod_type)
    cdef float noise_sqrd = noise_mag * noise_mag, real = 0, imag = 0, magnitude = 0, max_magnitude
    cdef float complex tmp

    if str(cython.typeof(samples)) == "char[:, ::1]":
        max_magnitude = sqrt(127*127 + 128*128)
    elif str(cython.typeof(samples)) == "unsigned char[:, ::1]":
        max_magnitude = sqrt(255*255)
    elif str(cython.typeof(samples)) == "short[:, ::1]":
        max_magnitude = sqrt(32768*32768 + 32767*32767)
    elif str(cython.typeof(samples)) == "unsigned short[:, ::1]":
        max_magnitude = sqrt(65535*65535)
    elif str(cython.typeof(samples)) == "float[:, ::1]":
        max_magnitude = sqrt(2)
    else:
        raise ValueError("Unsupported dtype")


    if mod_type == "PSK":
        return np.asarray(costa_demod(samples, noise_sqrd, mod_order, bandwidth=costas_loop_bandwidth))

    cdef float[::1] result = np.zeros(ns, dtype=np.float32, order="C")
    result[0] = NOISE

    for i in prange(1, ns, nogil=True, schedule="static"):
        real = samples[i, 0]
        imag = samples[i, 1]
        magnitude = real * real + imag * imag
        if magnitude <= noise_sqrd:  # |c| <= mag_treshold
            result[i] = NOISE
            continue

        if mod_type == "ASK":
            result[i] = sqrt(magnitude) / max_magnitude
        elif mod_type == "FSK":
            #tmp = samples[i - 1].conjugate() * c
            tmp = (samples[i-1, 0] - imag_unit * samples[i-1, 1]) * (real + imag_unit * imag)
            result[i] = atan2(tmp.imag, tmp.real)  # Freq

    return np.asarray(result)

cpdef np.ndarray[np.float32_t, ndim=1] get_center_thresholds(float center, float spacing, int modulation_order):
    cdef np.ndarray[np.float32_t, ndim=1] result = np.empty(modulation_order-1, dtype=np.float32)
    cdef int i, n = modulation_order // 2

    for i in range(0, n):
        result[i] = center - (n-(i+1)) * spacing

    for i in range(n, modulation_order-1):
        result[i] = center + (i+1-n) * spacing

    return result

cpdef int64_t[:, ::1] grab_pulse_lens(float[::1] samples, float center, uint16_t tolerance,
                                      str modulation_type, uint32_t samples_per_symbol,
                                      uint8_t bits_per_symbol=1, float center_spacing=0.1):
    """
    Get the pulse lengths after quadrature demodulation

    arr[i][0] gives type of symbol e.g. (arr[i][0] = 1) and (arr[i][0] = 0) for binary modulation
             Pause is (arr[i][0] = -1)
    arr[i][1] gives length of pulse
    """
    cdef bool is_ask = modulation_type == "ASK"
    cdef int64_t i, j, pulse_length = 0, num_samples = len(samples)
    cdef int64_t cur_index = 0, consecutive_ones = 0, consecutive_zeros = 0, consecutive_pause = 0
    cdef float s = 0, s_prev = 0
    cdef int cur_state = 0, new_state = 0, tmp_state = 0
    cdef float NOISE = get_noise_for_mod_type(modulation_type)

    cdef int modulation_order = 2**bits_per_symbol
    cdef int k


    cdef np.ndarray[np.float32_t, ndim=1] thresholds = get_center_thresholds(center, center_spacing, modulation_order)

    cdef int64_t[:, ::1] result = np.zeros((num_samples, 2), dtype=np.int64, order="C")
    if num_samples == 0:
        return result

    cdef int64_t[:] state_count = np.zeros(modulation_order, dtype=np.int64)

    s_prev = samples[0]
    if s_prev == NOISE:
        cur_state = PAUSE_STATE
    else:
        cur_state = modulation_order - 1
        for k in range(modulation_order - 1):
            if s <= thresholds[k]:
                cur_state = k
                break

    for i in range(num_samples):
        pulse_length += 1
        s = samples[i]

        if s == NOISE:
            tmp_state = PAUSE_STATE
        else:
            tmp_state = modulation_order - 1
            for k in range(modulation_order - 1):
                if s <= thresholds[k]:
                    tmp_state = k
                    break

        if tmp_state == PAUSE_STATE:
            consecutive_pause += 1
        else:
            consecutive_pause = 0

        for j in range(0, modulation_order):
            if j == tmp_state:
                state_count[j] += 1
            else:
                state_count[j] = 0

        if cur_state == tmp_state:
            continue

        new_state = -42

        if consecutive_pause > tolerance:
            new_state = PAUSE_STATE
        else:
            for j in range(0, modulation_order):
                if state_count[j] > tolerance:
                    new_state = j
                    break

        if new_state == -42:
            continue

        if is_ask and cur_state == PAUSE_STATE and (pulse_length - tolerance) < samples_per_symbol:
            # Aggregate short pauses for ASK
            cur_state = 0

        if cur_index > 0 and result[cur_index - 1, 0] == cur_state:
            result[cur_index - 1, 1] += pulse_length - tolerance
        else:
            result[cur_index, 0] = cur_state
            result[cur_index, 1] = pulse_length - tolerance
            cur_index += 1

        pulse_length = tolerance
        cur_state = new_state

    # Append last one
    cdef int64_t len_result = len(result)
    if cur_index < len_result:
        if cur_index > 0 and result[cur_index - 1, 0] == cur_state:
            result[cur_index - 1, 1] += pulse_length - tolerance
        else:
            result[cur_index, 0] = cur_state
            result[cur_index, 1] = pulse_length - tolerance
            cur_index += 1

    return result[:cur_index]

cpdef int find_nearest_center(float sample, float[::1] centers, int num_centers) nogil:
    cdef int i = 0
    cdef float center = 0
    cdef int result = 0
    cdef float min_diff = 99999
    cdef float cur_diff = 0

    for i in range(0, num_centers):
        center = centers[i]
        cur_diff = (sample - center) * (sample - center)
        if cur_diff < min_diff:
            min_diff = cur_diff
            result = i

    return result

cpdef np.ndarray[np.complex64_t, ndim=1] fir_filter(float complex[::1] input_samples, float complex[::1] filter_taps):
    cdef int i = 0, j = 0
    cdef int N = len(input_samples)
    cdef int M = len(filter_taps)
    cdef np.ndarray[np.complex64_t, ndim=1] output = np.zeros(N+M-1, dtype=np.complex64)


    for i in range(N):
        for j in range(M):
            output[i+j] += input_samples[i] * filter_taps[j]


    return output[:N]

cpdef np.ndarray[np.complex64_t, ndim=1] iir_filter(np.ndarray[np.float64_t, ndim=1] a,
                                                    np.ndarray[np.float64_t, ndim=1] b,
                                                    np.ndarray[np.complex64_t, ndim=1] signal):
    cdef np.ndarray[np.complex64_t, ndim=1] result = np.zeros(len(signal), dtype=np.complex64)

    cdef long n = 0, j = 0, k = 0
    cdef long M = len(a)
    cdef long N = len(b)
    for n in range(max(M, N+1) , len(signal)):
        for j in range(M):
            result[n] += a[j] * signal[n-j]

        for k in range(N):
            result[n] += b[k] * result[n-1-k]

    return result



from PyQt5.QtCore import pyqtSignal, Qt, pyqtSlot
from PyQt5.QtGui import QContextMenuEvent, QKeyEvent, QIcon
from PyQt5.QtWidgets import QListView, QMenu

from urh.models.GeneratorListModel import GeneratorListModel


class GeneratorListView(QListView):
    selection_changed = pyqtSignal()
    edit_on_item_triggered = pyqtSignal(int)

    def __init__(self, parent):
        super().__init__(parent)
        self.context_menu_pos = None

    def model(self) -> GeneratorListModel:
        return super().model()

    def create_context_menu(self):
        menu = QMenu()
        if self.model().message is None or len(self.model().message.message_type) == 0:
            return menu

        edit_action = menu.addAction("Edit fuzzing label")
        edit_action.setIcon(QIcon.fromTheme("configure"))
        edit_action.triggered.connect(self.on_edit_action_triggered)

        del_action = menu.addAction("Delete fuzzing label")
        del_action.setIcon(QIcon.fromTheme("edit-delete"))
        del_action.triggered.connect(self.on_delete_action_triggered)

        menu.addSeparator()
        fuzz_all_action = menu.addAction("Check all")
        fuzz_all_action.triggered.connect(self.model().fuzzAll)
        unfuzz_all_action = menu.addAction("Uncheck all")
        unfuzz_all_action.triggered.connect(self.model().unfuzzAll)

        return menu

    def contextMenuEvent(self, event: QContextMenuEvent):
        self.context_menu_pos = event.pos()
        menu = self.create_context_menu()
        menu.exec(self.mapToGlobal(self.context_menu_pos))
        self.context_menu_pos = None

    def selectionChanged(self, QItemSelection, QItemSelection_1):
        self.selection_changed.emit()
        super().selectionChanged(QItemSelection, QItemSelection_1)

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() in (Qt.Key_Enter, Qt.Key_Return):
            selected = [index.row() for index in self.selectedIndexes()]
            if len(selected) > 0:
                self.edit_on_item_triggered.emit(min(selected))
        else:
            super().keyPressEvent(event)

    def mouseDoubleClickEvent(self, QMouseEvent):
        selected = [index.row() for index in self.selectedIndexes()]
        if len(selected) > 0:
            self.edit_on_item_triggered.emit(min(selected))

    @pyqtSlot()
    def on_delete_action_triggered(self):
        index = self.indexAt(self.context_menu_pos)
        self.model().delete_label_at(index.row())

    @pyqtSlot()
    def on_edit_action_triggered(self):
        self.edit_on_item_triggered.emit(self.indexAt(self.context_menu_pos).row())



import math

from urh.signalprocessing.Signal import Signal
from urh.ui.painting.SceneManager import SceneManager


class SignalSceneManager(SceneManager):
    def __init__(self, signal: Signal, parent):
        super().__init__(parent)
        self.signal = signal
        self.scene_type = (
            0  # 0 = Analog Signal, 1 = QuadDemodView, 2 = Spectogram, 3 = I/Q view
        )
        self.mod_type = "ASK"

    def show_scene_section(
        self, x1: float, x2: float, subpath_ranges=None, colors=None
    ):
        if self.scene_type == 0:
            self.plot_data = self.signal.real_plot_data
        elif self.scene_type == 3:
            self.plot_data = [self.signal.imag_plot_data, self.signal.real_plot_data]
        else:
            self.plot_data = self.signal.qad
        super().show_scene_section(x1, x2, subpath_ranges=subpath_ranges, colors=colors)

    def init_scene(self):
        if self.scene_type == 0:
            # Ensure real plot has same y Axis
            self.plot_data = self.signal.real_plot_data
        elif self.scene_type == 3:
            self.plot_data = [self.signal.imag_plot_data, self.signal.real_plot_data]
        else:
            self.plot_data = self.signal.qad

        super().init_scene()
        if self.scene_type == 1 and (self.mod_type == "FSK" or self.mod_type == "PSK"):
            self.scene.setSceneRect(0, -4, self.num_samples, 8)

        self.line_item.setLine(0, 0, 0, 0)  # Hide Axis

        if self.scene_type == 0 or self.scene_type == 3:
            self.scene.draw_noise_area(
                self.signal.noise_min_plot,
                self.signal.noise_max_plot - self.signal.noise_min_plot,
            )
        else:
            self.scene.draw_sep_area(-self.signal.center_thresholds)

    def eliminate(self):
        super().eliminate()
        # do not eliminate the signal here, as it would cause data loss in tree models!
        # if hasattr(self.signal, "eliminate"):
        #    self.signal.eliminate()
        self.signal = None



from urh.signalprocessing.IQArray import IQArray
from urh.ui.painting.SceneManager import SceneManager


class LiveSceneManager(SceneManager):
    def __init__(self, data_array, parent):
        super().__init__(parent)
        self.plot_data = data_array
        self.end = 0

    @property
    def num_samples(self):
        return self.end



import os
import sys
import time

# from subprocess import PIPE, Popen
# from threading import Thread

import numpy as np
from PyQt5.QtCore import pyqtSignal

from urh import settings
from urh.plugins.Plugin import SDRPlugin
from urh.signalprocessing.Message import Message
from urh.util.Errors import Errors
from urh.util.Logger import logger


class FlipperZeroSubPlugin(SDRPlugin):
    def __init__(self):
        super().__init__(name="FlipperZeroSub")
        self.filetype = "Flipper SubGhz RAW File"
        self.version = 1
        self.protocol = "RAW"
        self.max_values_per_line = 512

    def getFuriHalString(self, modulation_type, given_bandwidth_deviation=0):
        if modulation_type == "ASK":
            if given_bandwidth_deviation > 500:
                # OOK, bandwidth 650kHz, asynchronous
                FuriHalString = "FuriHalSubGhzPresetOok650Async"
                bandwidth_deviation = 650
            else:
                # OOK, bandwidth 270kHz, asynchronous
                FuriHalString = "FuriHalSubGhzPresetOok270Async"
                bandwidth_deviation = 270
        elif modulation_type == "FSK":
            if given_bandwidth_deviation > 20:
                # FM, deviation 47.60742 kHz, asynchronous
                FuriHalString = "FuriHalSubGhzPreset2FSKDev476Async"
                bandwidth_deviation = 47.6
            else:
                # FM, deviation 2.380371 kHz, asynchronous
                FuriHalString = "FuriHalSubGhzPreset2FSKDev238Async"
                bandwidth_deviation = 2.38
        elif modulation_type == "GFSK":
            # GFSK, deviation 19.042969 kHz, 9.996Kb/s, asynchronous
            FuriHalString = "FuriHalSubGhzPresetGFSK9_99KbAsync"
            bandwidth_deviation = 19.04
        elif modulation_type == "PSK":
            FuriHalString = "FuriHalSubGhzPresetCustom"
            bandwidth_deviation = 238
        else:  # Fallback is Flipper Zero SubGHz default
            # OOK, bandwidth 650kHz, asynchronous
            FuriHalString = "FuriHalSubGhzPresetOok650Async"
            bandwidth_deviation = 650
        return FuriHalString, bandwidth_deviation

    def write_sub_file(
        self, filename, messages, sample_rates, modulators, project_manager
    ):
        # Check whether the message != NULL
        if len(messages) == 0:
            logger.debug("Empty signal!")
            return False

        # Open file "filename" for write
        try:
            file = open(filename, "w")
        except OSError as e:
            logger.debug(f"Could not open {filename} for writing: {e}", file=sys.stderr)
            return False

        # Get Flipper Zero FuriHal Preset and its datarate
        frequency = int(project_manager.device_conf["frequency"])
        # samplerate = sample_rates[0]
        samples_per_symbol = messages[0].samples_per_symbol
        preset, bandwidth_deviation = self.getFuriHalString(
            modulators[messages[0].modulator_index].modulation_type, 1000
        )  # TODO: last parameter is bandwidth/deviation of preset profile. Hardcoded to 1000 until it becomes clear how to find the value from signal data properly

        # Write Flipper Zero SubGHz protocol header
        file.write(f"Filetype: {self.filetype}\n")
        file.write(f"Version: {self.version}\n")
        file.write(f"Frequency: {frequency}\n")
        file.write(f"Preset: {preset}\n")
        file.write(f"Protocol: {self.protocol}")

        # Create signal
        signal = []
        for msg in messages:
            current_value = msg[0]
            current_count = 0
            for bit in msg:
                if bit == current_value:
                    current_count += 1
                else:
                    # Append to signal when bit changes
                    signal.append(
                        current_count if current_value == 1 else -current_count
                    )
                    current_count = 1
                    current_value = bit
            # Last pulse
            signal.append(current_count if current_value == 1 else -current_count)

        # Write signal
        sps = messages[0].samples_per_symbol
        for i in range(len(signal)):
            if 0 == i % self.max_values_per_line:
                file.write("\nRAW_Data:")
            file.write(f" {signal[i] * samples_per_symbol}")

        # Close file
        file.close()



import importlib
import os

from urh import settings
from urh.plugins.Plugin import Plugin, ProtocolPlugin
from urh.util.Logger import logger


class PluginManager(object):
    def __init__(self):
        self.plugin_path = os.path.dirname(os.path.realpath(__file__))
        self.installed_plugins = self.load_installed_plugins()

    @property
    def protocol_plugins(self):
        return [p for p in self.installed_plugins if isinstance(p, ProtocolPlugin)]

    def load_installed_plugins(self):
        """:rtype: list of Plugin"""
        result = []
        plugin_dirs = [
            d
            for d in os.listdir(self.plugin_path)
            if os.path.isdir(os.path.join(self.plugin_path, d))
        ]
        for d in plugin_dirs:
            if d == "__pycache__":
                continue
            try:
                class_module = self.load_plugin(d)
                plugin = class_module()
                plugin.plugin_path = os.path.join(self.plugin_path, plugin.name)
                plugin.load_description()
                if plugin.name in settings.all_keys():
                    plugin.enabled = settings.read(plugin.name, False, type=bool)
                else:
                    plugin.enabled = False
                result.append(plugin)
            except ImportError as e:
                logger.warning("Could not load plugin {0} ({1})".format(d, e))
                continue

        return result

    @staticmethod
    def load_plugin(plugin_name):
        classname = plugin_name + "Plugin"
        module_path = "urh.plugins." + plugin_name + "." + classname

        module = importlib.import_module(module_path)
        return getattr(module, classname)

    def is_plugin_enabled(self, plugin_name: str):
        return any(plugin_name == p.name for p in self.installed_plugins if p.enabled)

    def get_plugin_by_name(self, plugin_name):
        for plugin in self.installed_plugins:
            if plugin.name == plugin_name:
                return plugin
        return None



import operator
from enum import Enum
import xml.etree.ElementTree as ET

from urh.util.Logger import logger

OPERATIONS = {
    ">": operator.gt,
    "<": operator.lt,
    ">=": operator.ge,
    "<=": operator.le,
    "=": operator.eq,
    "!=": operator.ne,
}

OPERATION_DESCRIPTION = {
    ">": "greater",
    "<": "lower",
    ">=": "greater equal",
    "<=": "lower equal",
    "=": "equal",
    "!=": "not equal",
}


class Mode(Enum):
    all_apply = 0
    atleast_one_applies = 1
    none_applies = 2


class Rule(object):
    def __init__(
        self, start: int, end: int, operator: str, target_value: str, value_type: int
    ):
        assert operator in OPERATIONS
        self.__start = start
        self.__end = end + 1
        self.__value_type = value_type  # 0 = Bit, 1 = Hex, 2 = ASCII
        self.operator = operator
        self.target_value = target_value

    @property
    def start(self) -> int:
        return int(self.__start)

    @start.setter
    def start(self, value: int):
        try:
            self.__start = int(value)
        except ValueError:
            logger.warning("{} could not be cast to integer".format(value))

    @property
    def end(self) -> int:
        return int(self.__end)

    @end.setter
    def end(self, value: int):
        try:
            self.__end = int(value)
        except ValueError:
            logger.warning("{} could not be cast to integer".format(value))

    @property
    def value_type(self):
        return int(self.__value_type)

    @value_type.setter
    def value_type(self, value: int):
        try:
            self.__value_type = int(value)
        except ValueError:
            logger.warning("{} could not be cast to integer".format(value))

    def applies_for_message(self, message):
        data = (
            message.decoded_bits_str
            if self.value_type == 0
            else message.decoded_hex_str
            if self.value_type == 1
            else message.decoded_ascii_str
        )
        return OPERATIONS[self.operator](data[self.start : self.end], self.target_value)

    @property
    def operator_description(self):
        return OPERATION_DESCRIPTION[self.operator]

    @operator_description.setter
    def operator_description(self, value):
        for key, val in OPERATION_DESCRIPTION.items():
            if val == value:
                self.operator = key
                return
        logger.warning("Could not find operator description " + str(value))

    def to_xml(self) -> ET.Element:
        root = ET.Element("rule")

        for attr, val in vars(self).items():
            root.set(attr, str(val))

        return root

    @staticmethod
    def from_xml(tag: ET.Element):
        result = Rule(start=-1, end=-1, operator="=", target_value="", value_type=0)
        for attrib, value in tag.attrib.items():
            setattr(result, attrib, value)
        return result


class Ruleset(list):
    def __init__(self, mode: Mode = Mode.all_apply, rules=None):
        rules = rules if rules is not None else []
        self.mode = mode
        super().__init__(rules)

    def applies_for_message(self, message):
        napplied_rules = sum(rule.applies_for_message(message) for rule in self)

        if self.mode == Mode.all_apply:
            return napplied_rules == len(self)
        elif self.mode == Mode.atleast_one_applies:
            return napplied_rules > 0
        elif self.mode == Mode.none_applies:
            return napplied_rules == 0
        else:
            raise ValueError("Unknown behavior " + str(self.mode))

    def to_xml(self) -> ET.Element:
        root = ET.Element("ruleset")
        root.set("mode", str(self.mode.value))

        for rule in self:
            root.append(rule.to_xml())

        return root

    @staticmethod
    def from_xml(tag: ET.Element):
        if tag:
            result = Ruleset(mode=Mode(int(tag.get("mode", 0))))
            for rule in tag.findall("rule"):
                result.append(Rule.from_xml(rule))
            return result
        else:
            return Ruleset(mode=Mode.all_apply)



from PyQt5.QtCore import QAbstractTableModel, Qt, QModelIndex, pyqtSignal
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import qApp

from urh import settings
from urh.signalprocessing.FieldType import FieldType
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.simulator.SimulatorProtocolLabel import SimulatorProtocolLabel
from urh.util import util


class SimulatorMessageFieldModel(QAbstractTableModel):
    header_labels = ["Name", "Display format", "Value type", "Value"]

    protocol_label_updated = pyqtSignal(SimulatorProtocolLabel)

    def __init__(self, controller, parent=None):
        super().__init__(parent)

        self.controller = controller  # type: SimulatorTabController
        self.message_type = None  # type: MessageType

    def update(self):
        self.beginResetModel()
        self.message_type = self.controller.active_item.message_type
        self.endResetModel()

    def columnCount(self, parent: QModelIndex = None, *args, **kwargs):
        return len(self.header_labels)

    def rowCount(self, parent: QModelIndex = None, *args, **kwargs):
        return len(self.message_type) if self.message_type is not None else 0

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.header_labels[section]
        elif role == Qt.TextAlignmentRole:
            return Qt.AlignLeft

        return super().headerData(section, orientation, role)

    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        i, j = index.row(), index.column()
        lbl = self.message_type[i]  # type: SimulatorProtocolLabel

        if role == Qt.DisplayRole:
            if j == 0:
                return lbl.name
            elif j == 1:
                return ProtocolLabel.DISPLAY_FORMATS[lbl.display_format_index]
            elif j == 2:
                if lbl.is_checksum_label:
                    return "Checksum"
                else:
                    return lbl.VALUE_TYPES[lbl.value_type_index]
            elif j == 3:
                if lbl.value_type_index == 0:
                    message = lbl.parent()

                    try:
                        data = message.plain_bits[lbl.start : lbl.end]
                    except IndexError:
                        return None

                    return util.convert_bits_to_string(
                        data, lbl.display_format_index, pad_zeros=True
                    )
                elif lbl.value_type_index == 1:
                    return "-"
                elif lbl.value_type_index == 2:
                    return lbl.formula
                elif lbl.value_type_index == 3:
                    return lbl.external_program
                elif lbl.value_type_index == 4:
                    return (
                        "Range (Decimal): "
                        + str(lbl.random_min)
                        + " - "
                        + str(lbl.random_max)
                    )
        elif role == Qt.EditRole:
            if j == 0:
                return lbl.name
            elif j == 1:
                return lbl.display_format_index
            elif j == 2:
                return lbl.value_type_index
            elif j == 3:
                if lbl.value_type_index == 2:
                    return lbl.formula
                elif lbl.value_type_index == 3:
                    return lbl.external_program
                elif lbl.value_type_index == 4:
                    return [lbl.random_min, lbl.random_max]
        elif role == Qt.FontRole:
            if j == 0:
                font = QFont()
                font.setItalic(lbl.field_type is None)
                return font
            elif j == 2 and self.link_index(index):
                font = QFont()
                font.setUnderline(True)
                return font
        elif role == Qt.BackgroundColorRole:
            if j == 0:
                return settings.LABEL_COLORS[lbl.color_index]
            elif j == 3:
                if (
                    lbl.value_type_index == 2
                    and not self.controller.sim_expression_parser.validate_expression(
                        lbl.formula
                    )[0]
                ):
                    return settings.ERROR_BG_COLOR
        elif role == Qt.TextColorRole:
            if self.link_index(index):
                return qApp.palette().link().color()

    def link_index(self, index: QModelIndex):
        try:
            lbl = self.message_type[index.row()]  # type: SimulatorProtocolLabel
            if index.column() == 2 and lbl.is_checksum_label:
                return True
        except:
            return False
        return False

    def setData(self, index: QModelIndex, value, role=None):
        if role == Qt.EditRole:
            i, j = index.row(), index.column()
            label = self.message_type[i]  # type: SimulatorProtocolLabel

            if j == 0:
                label.name = value
                ft = self.controller.field_types_by_caption.get(
                    value, FieldType("Custom", FieldType.Function.CUSTOM)
                )
                label.field_type = ft
            elif j == 1:
                label.display_format_index = value
            elif j == 2:
                label.value_type_index = value
            elif j == 3:
                if label.value_type_index == 0:
                    message = label.parent()
                    try:
                        bits = util.convert_string_to_bits(
                            value,
                            label.display_format_index,
                            target_num_bits=label.end - label.start,
                        )

                        message.plain_bits[label.start : label.end] = bits
                    except ValueError:
                        pass
                elif label.value_type_index == 2:
                    label.formula = value
                elif label.value_type_index == 3:
                    label.external_program = value
                elif label.value_type_index == 4:
                    label.random_min = value[0]
                    label.random_max = value[1]
            self.dataChanged.emit(self.index(i, 0), self.index(i, self.columnCount()))
            self.protocol_label_updated.emit(label)

        return True

    def flags(self, index: QModelIndex):
        row, col = index.row(), index.column()
        label = self.message_type[row]  # type: SimulatorProtocolLabel

        if col == 2 and label.is_checksum_label:
            return Qt.ItemIsSelectable

        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable

        if not (col == 3 and label.value_type_index == 1):
            flags |= Qt.ItemIsEditable

        return flags

    def remove_label_at(self, index: int):
        try:
            label = self.message_type[index]
            self.controller.simulator_config.delete_items([label])
        except IndexError:
            pass

    def set_value_type_index(self, rows: list, value_type_index: int):
        for row in rows:
            label = self.message_type[row]
            if not label.is_checksum_label:
                label.value_type_index = value_type_index
                self.protocol_label_updated.emit(label)
        self.update()



cimport urh.dev.native.lib.csdrplay as csdrplay
import time
from libc.stdlib cimport malloc, free

ctypedef csdrplay.mir_sdr_DeviceT device_type
ctypedef csdrplay.mir_sdr_ErrT error_t

cdef extern from "Python.h":
    ctypedef enum PyGILState_STATE:
        PyGILState_LOCKED
        PyGILState_UNLOCKED
    PyGILState_STATE PyGILState_Ensure()
    void PyGILState_Release(PyGILState_STATE)



cdef csdrplay.mir_sdr_SetGrModeT set_gr_mode=csdrplay.mir_sdr_USE_RSP_SET_GR

global reset_rx, reset_rx_request_received
reset_rx = False
reset_rx_request_received = False

cdef void __rx_stream_callback(short *xi, short *xq, unsigned int firstSampleNum, int grChanged, int rfChanged, int fsChanged, unsigned int numSamples, unsigned int reset, void *cbContext):
    cdef short* data = <short *>malloc(2*numSamples * sizeof(short))

    cdef unsigned int i = 0
    cdef unsigned int j = 0

    global reset_rx, reset_rx_request_received
    if reset_rx:
        reset_rx_request_received = True
        return

    cdef PyGILState_STATE gstate

    try:
        for i in range(0, numSamples):
            data[j] = xi[i]
            data[j+1] = xq[i]
            j += 2

        gstate = PyGILState_Ensure()
        conn = <object> cbContext
        conn.send_bytes(<short[:2*numSamples]>data)  # python callback
        return
    finally:
        PyGILState_Release(gstate)
        free(data)

cdef void _rx_stream_callback(short *xi, short *xq, unsigned int firstSampleNum, int grChanged, int rfChanged, int fsChanged, unsigned int numSamples, unsigned int reset, unsigned int hwRemoved, void *cbContext) noexcept:
    __rx_stream_callback(xi, xq, firstSampleNum, grChanged, rfChanged, fsChanged, numSamples, reset, cbContext)

cdef void _gain_change_callback(unsigned int gRdB, unsigned int lnaGRdB, void *cbContext) noexcept:
    return

cpdef void set_gr_mode_for_dev_model(int dev_model):
    global set_gr_mode
    if dev_model == 1:
        set_gr_mode = csdrplay.mir_sdr_USE_SET_GR_ALT_MODE
    else:
        set_gr_mode = csdrplay.mir_sdr_USE_RSP_SET_GR

cpdef float get_api_version():
    cdef float version = 0.0
    csdrplay.mir_sdr_ApiVersion(&version)
    return version

cpdef error_t set_device_index(unsigned int index):
    return csdrplay.mir_sdr_SetDeviceIdx(index)

cpdef error_t release_device_index():
    return csdrplay.mir_sdr_ReleaseDeviceIdx()

cpdef get_devices():
    cdef device_type *devs = <device_type*> malloc(256 * sizeof(device_type))

    if not devs:
        raise MemoryError()

    cdef unsigned int num_devs = 0
    try:
        csdrplay.mir_sdr_GetDevices(devs, &num_devs, 256)

        result = []

        for i in range(num_devs):
            d = {"serial": devs[i].SerNo.decode("iso-8859-1"), "device_ref": devs[i].DevNm.decode("iso-8859-1"),
                 "hw_version": devs[i].hwVer, "available": devs[i].devAvail}
            result.append(d)

        return result
    finally:
        free(devs)

cdef int calculate_gain_reduction(int gain):
    """
    Calculate gain reduction for API. Highest possible gain leads to lowest possible gain reduction
    :param gain:
    :return:
    """
    gain = max(20, min(gain, 59))
    return 79 - gain

cdef csdrplay.mir_sdr_Bw_MHzT get_nearest_bandwidth(double bandwidth):
    # get nearest bwtype
    bandwidths = {200e3: csdrplay.mir_sdr_BW_0_200,
                  300e3: csdrplay.mir_sdr_BW_0_300,
                  600e3: csdrplay.mir_sdr_BW_0_600,
                  1536e3: csdrplay.mir_sdr_BW_1_536,
                  5000e3: csdrplay.mir_sdr_BW_5_000,
                  6000e3: csdrplay.mir_sdr_BW_6_000,
                  7000e3: csdrplay.mir_sdr_BW_7_000,
                  8000e3: csdrplay.mir_sdr_BW_8_000}

    cdef csdrplay.mir_sdr_Bw_MHzT bw_type = csdrplay.mir_sdr_Bw_MHzT.mir_sdr_BW_0_200
    best_match = 0
    for bw in bandwidths:
        if abs(bw - bandwidth) < abs(best_match - bandwidth):
            best_match = bw
    bw_type = bandwidths[best_match]
    return bw_type

cdef csdrplay.mir_sdr_If_kHzT get_nearest_if_gain(double if_gain):
    cdef csdrplay.mir_sdr_If_kHzT if_type = csdrplay.mir_sdr_If_kHzT.mir_sdr_IF_Zero
    best_match = 0
    if_types = {0: csdrplay.mir_sdr_IF_Zero,
                450: csdrplay.mir_sdr_IF_0_450,
                1620: csdrplay.mir_sdr_IF_1_620,
                2048: csdrplay.mir_sdr_IF_2_048}
    for i in if_types:
        if abs(i - if_gain) < abs(best_match - if_gain):
            best_match = i

    if_type = if_types[best_match]
    return if_type

cpdef init_stream(int gain, double sample_rate, double center_freq, double bandwidth, double if_gain, object func):
    global set_gr_mode

    cdef csdrplay.mir_sdr_Bw_MHzT bw_type = get_nearest_bandwidth(bandwidth)
    # get nearest ifgain
    cdef csdrplay.mir_sdr_If_kHzT if_type = get_nearest_if_gain(if_gain)

    lna_state = 0
    cdef int gRdBsystem = 0
    cdef int samples_per_packet = 0

    cdef int gain_reduction = calculate_gain_reduction(gain)
    return csdrplay.mir_sdr_StreamInit(&gain_reduction, sample_rate / 1e6, center_freq / 1e6, bw_type, if_type, lna_state,
                                       &gRdBsystem, set_gr_mode, &samples_per_packet, _rx_stream_callback,
                                       _gain_change_callback, <void *> func)

cpdef error_t set_center_freq(double frequency):
    return reinit_stream(csdrplay.mir_sdr_CHANGE_RF_FREQ, frequency=frequency)

cpdef error_t set_sample_rate(double sample_rate):
    return reinit_stream(csdrplay.mir_sdr_CHANGE_FS_FREQ, sample_rate=sample_rate)

cpdef error_t set_bandwidth(double bandwidth):
    cdef csdrplay.mir_sdr_Bw_MHzT bw_type = get_nearest_bandwidth(bandwidth)
    return reinit_stream(csdrplay.mir_sdr_CHANGE_BW_TYPE, bw_type=bw_type)

cpdef error_t set_gain(int gain):
    return reinit_stream(csdrplay.mir_sdr_CHANGE_GR, gain=calculate_gain_reduction(gain))

cpdef error_t set_if_gain(double if_gain):
    cdef csdrplay.mir_sdr_If_kHzT if_type = get_nearest_if_gain(if_gain)
    return reinit_stream(csdrplay.mir_sdr_CHANGE_IF_TYPE, if_type=if_type)

cpdef error_t set_antenna(int antenna):
    cdef csdrplay.mir_sdr_RSPII_AntennaSelectT antenna_select
    if antenna == 0 or antenna == 1:
        result = csdrplay.mir_sdr_AmPortSelect(0)
        if result != csdrplay.mir_sdr_Success:
            return result

        if antenna == 0:
            antenna_select = csdrplay.mir_sdr_RSPII_ANTENNA_A
        else:
            antenna_select = csdrplay.mir_sdr_RSPII_ANTENNA_B

        return csdrplay.mir_sdr_RSPII_AntennaControl(antenna_select)
    elif antenna == 2:
        print("hiz")
        return csdrplay.mir_sdr_AmPortSelect(1)


cpdef error_t reinit_stream(csdrplay.mir_sdr_ReasonForReinitT reason_for_reinit,
                            double sample_rate=0, double frequency=0,
                            csdrplay.mir_sdr_Bw_MHzT bw_type=csdrplay.mir_sdr_BW_Undefined,
                            int gain=0,
                            csdrplay.mir_sdr_If_kHzT if_type=csdrplay.mir_sdr_IF_Undefined,
                            csdrplay.mir_sdr_LoModeT lo_mode=csdrplay.mir_sdr_LO_Undefined,
                            int lna_state=0):
    cdef int gRdBsystem, samplesPerPacket
    global reset_rx, reset_rx_request_received, set_gr_mode
    reset_rx = True

    while not reset_rx_request_received:
        time.sleep(0.01)

    try:
        return csdrplay.mir_sdr_Reinit(&gain, sample_rate / 1e6, frequency / 1e6, bw_type, if_type, lo_mode, lna_state, &gRdBsystem, set_gr_mode, &samplesPerPacket, reason_for_reinit)
    finally:
        reset_rx = False
        reset_rx_request_received = False

cpdef error_t close_stream():
    csdrplay.mir_sdr_StreamUninit()



from PyQt5.QtCore import QAbstractTableModel, pyqtSignal, Qt, QModelIndex

from urh.signalprocessing.Message import Message
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.simulator.SimulatorProtocolLabel import SimulatorProtocolLabel


class PLabelTableModel(QAbstractTableModel):
    header_labels = ["Name", "Start", "End", "Color", "Apply decoding"]

    label_removed = pyqtSignal(ProtocolLabel)
    special_status_label_changed = pyqtSignal(ProtocolLabel)
    apply_decoding_changed = pyqtSignal(ProtocolLabel)

    def __init__(self, message: Message, field_types, parent=None):
        """

        :param message:
        :type field_types: list of FieldType
        :param parent:
        """
        super().__init__(parent)
        self.row_count = len(message.message_type)
        self.proto_view = 0
        self.__message = None
        self.message = message

        self.message_type = message.message_type  # type: MessageType
        self.field_types_by_caption = {ft.caption: ft for ft in field_types}
        self.update()

    @property
    def message(self) -> Message:
        return self.__message

    @message.setter
    def message(self, value: Message):
        self.__message = value
        # Ensure bit alignment positions in message are set
        self.__message.split(decode=True)

    def __get_label_at(self, index: int) -> ProtocolLabel:
        result = self.message_type[index]
        if isinstance(result, SimulatorProtocolLabel):
            return result.label
        else:
            return result

    def update(self):
        self.beginResetModel()
        self.endResetModel()

    def columnCount(self, parent: QModelIndex = None, *args, **kwargs):
        return len(self.header_labels)

    def rowCount(self, parent: QModelIndex = None, *args, **kwargs):
        return len(self.message_type)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.header_labels[section]
        return super().headerData(section, orientation, role)

    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        i, j = index.row(), index.column()
        if role == Qt.DisplayRole:
            try:
                lbl = self.message_type[i]
            except IndexError:
                return False
            if j == 0:
                return lbl.name
            elif j == 1:
                return (
                    self.message.get_label_range(
                        lbl, view=self.proto_view, decode=True
                    )[0]
                    + 1
                )
            elif j == 2:
                return self.message.get_label_range(
                    lbl, view=self.proto_view, decode=True
                )[1]
            elif j == 3:
                return lbl.color_index
            elif j == 4:
                return lbl.apply_decoding
        elif role == Qt.TextAlignmentRole:
            return Qt.AlignCenter
        else:
            return None

    def setData(self, index: QModelIndex, value, role=Qt.EditRole):
        if value == "":
            return True

        i = index.row()
        j = index.column()
        if i >= len(self.message_type):
            return False

        lbl = self.__get_label_at(i)

        if j == 0:
            lbl.name = value
            type_before = type(lbl)
            self.message_type.change_field_type_of_label(
                lbl, self.field_types_by_caption.get(value, None)
            )

            lbl = self.__get_label_at(i)

            if type_before != ProtocolLabel or type(lbl) != ProtocolLabel:
                self.special_status_label_changed.emit(lbl)

        elif j == 1:
            lbl.start = self.message.convert_index(
                int(value - 1), from_view=self.proto_view, to_view=0, decoded=True
            )[0]
        elif j == 2:
            lbl.end = self.message.convert_index(
                int(value), from_view=self.proto_view, to_view=0, decoded=True
            )[0]
        elif j == 3:
            lbl.color_index = value
        elif j == 4:
            if bool(value) != lbl.apply_decoding:
                lbl.apply_decoding = bool(value)
                self.apply_decoding_changed.emit(lbl)

        return True

    def flags(self, index):
        if not index.isValid():
            return Qt.NoItemFlags

        try:
            _ = self.message_type[index.row()]
        except IndexError:
            return Qt.NoItemFlags

        return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def remove_label(self, label):
        self.message_type.remove(label)
        self.update()
        self.label_removed.emit(label)

    def remove_label_at(self, index: int):
        try:
            label = self.message_type[index]
            self.remove_label(label)
        except IndexError:
            pass



import copy

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QDropEvent
from PyQt5.QtWidgets import (
    QGraphicsScene,
    QGraphicsSceneDragDropEvent,
    QAbstractItemView,
)

from urh.signalprocessing.FieldType import FieldType
from urh.signalprocessing.Message import Message
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.Participant import Participant
from urh.simulator.ActionItem import (
    ActionItem,
    GotoActionItem,
    TriggerCommandActionItem,
    SleepActionItem,
    CounterActionItem,
)
from urh.simulator.GraphicsItem import GraphicsItem
from urh.simulator.LabelItem import LabelItem
from urh.simulator.MessageItem import MessageItem
from urh.simulator.ParticipantItem import ParticipantItem
from urh.simulator.RuleItem import RuleItem, RuleConditionItem
from urh.simulator.SimulatorConfiguration import SimulatorConfiguration
from urh.simulator.SimulatorCounterAction import SimulatorCounterAction
from urh.simulator.SimulatorGotoAction import SimulatorGotoAction
from urh.simulator.SimulatorItem import SimulatorItem
from urh.simulator.SimulatorMessage import SimulatorMessage
from urh.simulator.SimulatorProtocolLabel import SimulatorProtocolLabel
from urh.simulator.SimulatorRule import (
    SimulatorRule,
    SimulatorRuleCondition,
    ConditionType,
)
from urh.simulator.SimulatorSleepAction import SimulatorSleepAction
from urh.simulator.SimulatorTriggerCommandAction import SimulatorTriggerCommandAction


class SimulatorScene(QGraphicsScene):
    files_dropped = pyqtSignal(list)

    model_to_scene_class_mapping = {
        SimulatorRule: RuleItem,
        SimulatorRuleCondition: RuleConditionItem,
        SimulatorGotoAction: GotoActionItem,
        SimulatorTriggerCommandAction: TriggerCommandActionItem,
        SimulatorCounterAction: CounterActionItem,
        SimulatorSleepAction: SleepActionItem,
        SimulatorMessage: MessageItem,
        SimulatorProtocolLabel: LabelItem,
    }

    def __init__(
        self, mode: int, simulator_config: SimulatorConfiguration, parent=None
    ):
        super().__init__(parent)
        self.mode = mode
        self.simulator_config = simulator_config
        self.tree_root_item = None

        self.participants_dict = {}
        self.participant_items = []

        self.broadcast_part = self.insert_participant(
            self.simulator_config.broadcast_part
        )
        self.not_assigned_part = self.insert_participant(None)
        self.update_participants(refresh=False)

        self.items_dict = {}

        self.on_items_added([item for item in self.simulator_config.rootItem.children])

        self.create_connects()

    @property
    def visible_participants(self):
        return [part for part in self.participant_items if part.isVisible()]

    @property
    def visible_participants_without_broadcast(self):
        return [
            part
            for part in self.participant_items
            if part.isVisible() and part is not self.broadcast_part
        ]

    def create_connects(self):
        self.simulator_config.participants_changed.connect(self.update_participants)

        self.simulator_config.items_deleted.connect(self.on_items_deleted)
        self.simulator_config.items_updated.connect(self.on_items_updated)
        self.simulator_config.items_moved.connect(self.on_items_moved)
        self.simulator_config.items_added.connect(self.on_items_added)

    def on_items_deleted(self, items):
        if self is None:
            return

        for item in items:
            scene_item = self.model_to_scene(item)

            if scene_item in self.items():
                self.removeItem(scene_item)

        self.update_items_dict()
        self.update_view()

    def on_items_updated(self, items):
        scene_items = [self.model_to_scene(item) for item in items]

        for scene_item in scene_items:
            scene_item.refresh()

        self.update_view()

    def on_items_moved(self, items):
        for item in items:
            scene_item = self.model_to_scene(item)
            self.insert_item(scene_item)

        self.update_view()

    def on_items_added(self, items):
        for item in items:
            self.on_item_added(item)

        self.update_view()

    def on_item_added(self, item: SimulatorItem):
        scene_item = self.model_to_scene_class_mapping[type(item)](model_item=item)
        self.insert_item(scene_item)

        # add children to scene ...
        for child in item.children:
            self.on_item_added(child)

    def model_to_scene(self, model_item: SimulatorItem):
        if model_item is None or model_item is self.simulator_config.rootItem:
            return None

        try:
            return self.items_dict[model_item]
        except KeyError:
            return None

    def insert_participant(self, participant: Participant):
        participant_item = ParticipantItem(participant)
        participant_item.setVisible(False)
        self.participant_items.insert(-2, participant_item)
        self.participants_dict[participant] = participant_item
        self.addItem(participant_item)

        return participant_item

    def insert_item(self, item: GraphicsItem):
        parent_scene_item = self.get_parent_scene_item(item)
        item.setParentItem(parent_scene_item)

        self.items_dict[item.model_item] = item

        if item not in self.items():
            self.addItem(item)

        item.update_flags()
        item.refresh()

    def get_parent_scene_item(self, item: GraphicsItem):
        return self.model_to_scene(item.model_item.parent())

    def min_items_width(self):
        width = 0
        items = [
            item
            for item in self.items()
            if isinstance(item, (RuleConditionItem, ActionItem))
        ]

        for item in items:
            if item.labels_width() > width:
                width = item.labels_width()

        return width

    def items_width(self):
        vp = self.visible_participants

        if len(vp) >= 2:
            width = vp[-1].x_pos()
            width -= vp[0].x_pos()
        else:
            width = self.min_items_width()

        return width

    def delete_selected_items(self):
        items = self.selectedItems()
        self.clearSelection()

        self.simulator_config.delete_items([item.model_item for item in items])

    def log_selected_items(self, logging_active: bool):
        items = self.selectedItems()
        self.log_items(items, logging_active)

    def log_items(self, items, logging_active: bool):
        for item in items:
            item.model_item.logging_active = logging_active

        self.simulator_config.items_updated.emit([item.model_item for item in items])

    def log_toggle_selected_items(self):
        items = self.selectedItems()

        for item in items:
            item.model_item.logging_active = not item.model_item.logging_active

        self.simulator_config.items_updated.emit([item.model_item for item in items])

    def log_all_items(self, logging_active: bool):
        self.log_items(self.selectable_items(), logging_active)

    def selectable_items(self):
        return [
            item
            for item in self.items()
            if isinstance(item, GraphicsItem) and item.is_selectable()
        ]

    def move_items(self, items, ref_item, position):
        new_pos, new_parent = self.insert_at(ref_item, position)
        self.simulator_config.move_items(items, new_pos, new_parent)

    def select_all_items(self):
        for item in self.simulator_config.rootItem.children:
            scene_item = self.model_to_scene(item)
            scene_item.select_all()

    def update_numbering(self):
        for item in self.simulator_config.rootItem.children:
            scene_item = self.model_to_scene(item)
            scene_item.update_numbering()

    def update_valid_states(self):
        self.simulator_config.update_valid_states()

    def update_view(self):
        self.update_numbering()
        self.update_valid_states()
        self.arrange_participants()
        self.arrange_items()

        # resize scrollbar
        self.setSceneRect(self.itemsBoundingRect().adjusted(-10, 0, 0, 0))

    def update_participants(self, refresh=True):
        for participant in list(self.participants_dict):
            if (
                participant is None
                or participant == self.simulator_config.broadcast_part
            ):
                continue

            self.removeItem(self.participants_dict[participant])
            self.participant_items.remove(self.participants_dict[participant])
            del self.participants_dict[participant]

        for participant in self.simulator_config.participants:
            if participant in self.participants_dict:
                self.participants_dict[participant].refresh()
            else:
                self.insert_participant(participant)

        if refresh:
            self.update_view()

    def refresh_participant(self, participant: Participant):
        try:
            self.participants_dict[participant].refresh()
        except KeyError:
            pass

    def update_items_dict(self):
        sim_items = self.simulator_config.get_all_items()

        for key in list(self.items_dict.keys()):
            if key not in sim_items:
                del self.items_dict[key]

    def get_all_message_items(self):
        """

        :rtype: list[MessageItem]
        """
        return [item for item in self.items() if isinstance(item, MessageItem)]

    def get_selected_messages(self):
        """

        :rtype: list[SimulatorMessage]
        """
        return [
            item.model_item
            for item in self.selectedItems()
            if isinstance(item, MessageItem)
        ]

    def select_messages_with_participant(
        self, participant: ParticipantItem, from_part=True
    ):
        messages = self.get_all_message_items()
        self.clearSelection()

        for msg in messages:
            if (from_part and msg.source is participant) or (
                not from_part and msg.destination is participant
            ):
                msg.select_all()

    def arrange_participants(self):
        messages = self.get_all_message_items()

        for participant in self.participant_items:
            if any(
                msg.source == participant or msg.destination == participant
                for msg in messages
            ):
                participant.setVisible(True)
            else:
                participant.setVisible(False)
                participant.update_position(x_pos=30)

        vp = self.visible_participants

        if not vp:
            return

        vp[0].update_position(x_pos=0)

        for i in range(1, len(vp)):
            curr_participant = vp[i]
            participants_left = vp[:i]

            items = [
                msg
                for msg in messages
                if (
                    (
                        msg.source == curr_participant
                        and msg.destination in participants_left
                    )
                    or (
                        msg.source in participants_left
                        and msg.destination == curr_participant
                    )
                )
            ]

            x_max = vp[i - 1].x_pos()
            x_max += (vp[i - 1].width() + curr_participant.width()) / 2
            x_max += 10

            for msg in items:
                x = msg.width() + 30
                x += (
                    msg.source.x_pos()
                    if msg.source != curr_participant
                    else msg.destination.x_pos()
                )

                if x > x_max:
                    x_max = x

            if i == len(vp) - 1:
                if self.min_items_width() > x_max:
                    x_max = self.min_items_width()

            curr_participant.update_position(x_pos=x_max)

    def arrange_items(self):
        x_pos = 0
        y_pos = 30

        for item in self.simulator_config.rootItem.children:
            scene_item = self.model_to_scene(item)
            scene_item.update_position(x_pos, y_pos)
            y_pos += round(scene_item.boundingRect().height())

        for participant in self.participant_items:
            participant.update_position(y_pos=max(y_pos, 50))

    def dragMoveEvent(self, event: QGraphicsSceneDragDropEvent):
        if any(item.acceptDrops() for item in self.items(event.scenePos())):
            super().dragMoveEvent(event)
        else:
            event.setAccepted(True)

    def insert_at(self, ref_item, position, insert_rule=False):
        if ref_item:
            ref_item = ref_item.model_item

        if ref_item is None:
            parent_item = self.simulator_config.rootItem
            insert_position = self.simulator_config.n_top_level_items()
        elif insert_rule:
            parent_item = self.simulator_config.rootItem

            while ref_item.parent() != self.simulator_config.rootItem:
                ref_item = ref_item.parent()

            insert_position = ref_item.get_pos()
        elif isinstance(ref_item, SimulatorRuleCondition):
            if position == QAbstractItemView.OnItem:
                parent_item = ref_item
                insert_position = parent_item.child_count()
            else:
                parent_item = self.simulator_config.rootItem
                insert_position = ref_item.parent().get_pos()
        else:
            parent_item = ref_item.parent()
            insert_position = ref_item.get_pos()

        if position == QAbstractItemView.BelowItem:
            insert_position += 1

        return (insert_position, parent_item)

    def dropEvent(self, event: QDropEvent):
        items = [
            item
            for item in self.items(event.scenePos())
            if isinstance(item, GraphicsItem) and item.acceptDrops()
        ]
        item = None if len(items) == 0 else items[0]
        if len(event.mimeData().urls()) > 0:
            self.files_dropped.emit(event.mimeData().urls())

        indexes = list(event.mimeData().text().split("/")[:-1])

        group_nodes = []
        file_nodes = []
        for index in indexes:
            try:
                row, column, parent = map(int, index.split(","))
                if parent == -1:
                    parent = self.tree_root_item
                else:
                    parent = self.tree_root_item.child(parent)
                node = parent.child(row)
                if node.is_group:
                    group_nodes.append(node)
                else:
                    file_nodes.append(node)
            except ValueError:
                continue

        # Which Nodes to add?
        nodes_to_add = []
        """:type: list of ProtocolTreeItem """
        for group_node in group_nodes:
            nodes_to_add.extend(group_node.children)
        nodes_to_add.extend(
            [file_node for file_node in file_nodes if file_node not in nodes_to_add]
        )
        protocols_to_add = [node.protocol for node in nodes_to_add]

        ref_item = item
        position = None if ref_item is None else item.drop_indicator_position
        self.add_protocols(ref_item, position, protocols_to_add)
        super().dropEvent(event)

    def add_rule(self, ref_item, position):
        rule = SimulatorRule()
        pos, parent = self.insert_at(ref_item, position, True)
        self.simulator_config.add_items([rule], pos, parent)

        self.add_rule_condition(rule, ConditionType.IF)
        return rule

    def add_rule_condition(self, rule: SimulatorRule, type: ConditionType):
        rule_condition = SimulatorRuleCondition(type)

        pos = rule.child_count()

        if type is ConditionType.ELSE_IF and rule.has_else_condition:
            pos -= 1

        self.simulator_config.add_items([rule_condition], pos, rule)
        return rule_condition

    def add_goto_action(self, ref_item, position):
        goto_action = SimulatorGotoAction()
        pos, parent = self.insert_at(ref_item, position, False)
        self.simulator_config.add_items([goto_action], pos, parent)
        return goto_action

    def add_sleep_action(self, ref_item, position):
        sleep_action = SimulatorSleepAction()
        pos, parent = self.insert_at(ref_item, position, False)
        self.simulator_config.add_items([sleep_action], pos, parent)
        return sleep_action

    def add_counter_action(self, ref_item, position):
        counter_action = SimulatorCounterAction()
        pos, parent = self.insert_at(ref_item, position, False)
        self.simulator_config.add_items([counter_action], pos, parent)
        return counter_action

    def add_trigger_command_action(self, ref_item, position):
        command_action = SimulatorTriggerCommandAction()
        pos, parent = self.insert_at(ref_item, position, False)
        self.simulator_config.add_items([command_action], pos, parent)
        return command_action

    def add_message(
        self,
        plain_bits,
        pause,
        message_type,
        ref_item,
        position,
        decoder=None,
        source=None,
        destination=None,
    ):
        message = self.create_message(
            destination, plain_bits, pause, message_type, decoder, source
        )
        pos, parent = self.insert_at(ref_item, position, False)
        self.simulator_config.add_items([message], pos, parent)
        return message

    def create_message(
        self, destination, plain_bits, pause, message_type, decoder, source
    ):
        if destination is None:
            destination = self.simulator_config.broadcast_part

        sim_message = SimulatorMessage(
            destination=destination,
            plain_bits=plain_bits,
            pause=pause,
            message_type=MessageType(message_type.name),
            decoder=decoder,
            source=source,
        )

        for lbl in message_type:
            sim_label = SimulatorProtocolLabel(copy.deepcopy(lbl))
            sim_message.insert_child(-1, sim_label)

        return sim_message

    def clear_all(self):
        self.simulator_config.delete_items(
            [item for item in self.simulator_config.rootItem.children]
        )

    def add_protocols(self, ref_item, position, protocols_to_add: list):
        pos, parent = self.insert_at(ref_item, position)
        messages = []

        for protocol in protocols_to_add:
            for msg in protocol.messages:
                source, destination = self.detect_source_destination(msg)
                simulator_msg = self.create_message(
                    destination=destination,
                    plain_bits=copy.copy(msg.decoded_bits),
                    pause=0,
                    message_type=msg.message_type,
                    decoder=msg.decoder,
                    source=source,
                )
                simulator_msg.timestamp = msg.timestamp
                messages.append(simulator_msg)

        self.simulator_config.add_items(messages, pos, parent)

    def get_drag_nodes(self):
        drag_nodes = []
        self.__get_drag_nodes(self.simulator_config.rootItem, drag_nodes)
        return drag_nodes

    def __get_drag_nodes(self, node: SimulatorItem, drag_nodes: list):
        scene_item = self.model_to_scene(node)

        if scene_item and scene_item.isSelected() and scene_item.is_movable():
            drag_nodes.append(scene_item.model_item)

        for child in node.children:
            self.__get_drag_nodes(child, drag_nodes)

    def detect_source_destination(self, message: Message):
        participants = self.simulator_config.participants

        source = None if len(participants) < 2 else participants[0]
        destination = self.simulator_config.broadcast_part

        if message.participant:
            source = message.participant
            dst_address_label = next(
                (
                    lbl
                    for lbl in message.message_type
                    if lbl.field_type
                    and lbl.field_type.function == FieldType.Function.DST_ADDRESS
                ),
                None,
            )
            if dst_address_label:
                start, end = message.get_label_range(
                    dst_address_label, view=1, decode=True
                )
                dst_address = message.decoded_hex_str[start:end]
                dst = next(
                    (p for p in participants if p.address_hex == dst_address), None
                )
                if dst is not None and dst != source:
                    destination = dst

        return source, destination



import xml.etree.ElementTree as ET
from enum import Enum

from urh.simulator.SimulatorItem import SimulatorItem


class SimulatorRule(SimulatorItem):
    def __init__(self):
        super().__init__()

    def set_parent(self, value):
        if value is not None:
            assert value.parent() is None

        super().set_parent(value)

    @property
    def has_else_condition(self) -> bool:
        return any(child.type is ConditionType.ELSE for child in self.children)

    def get_first_applying_condition(self):
        return next((child for child in self.children if child.condition_applies), None)

    def next_item(self):
        return next(
            (
                c.children[0]
                for c in self.children
                if c.condition_applies and c.child_count()
            ),
            self.next_sibling(),
        )

    def to_xml(self) -> ET.Element:
        return ET.Element("simulator_rule")

    @classmethod
    def from_xml(cls, tag: ET.Element):
        return SimulatorRule()


class ConditionType(Enum):
    IF = "IF"
    ELSE_IF = "ELSE IF"
    ELSE = "ELSE"


class SimulatorRuleCondition(SimulatorItem):
    def __init__(self, type: ConditionType):
        super().__init__()
        self.type = type
        self.condition = ""

    @property
    def condition_applies(self) -> bool:
        if self.type is ConditionType.ELSE:
            return True

        valid, _, node = self.expression_parser.validate_expression(
            self.condition, is_formula=False
        )
        assert valid == True and node is not None
        return self.expression_parser.evaluate_node(node)

    def set_parent(self, value):
        if value is not None:
            assert isinstance(value, SimulatorRule)

        super().set_parent(value)

    def validate(self):
        if self.type is ConditionType.ELSE:
            return True

        result, _, _ = self.expression_parser.validate_expression(
            self.condition, is_formula=False
        )
        return result

    def to_xml(self):
        return ET.Element(
            "simulator_rule_condition",
            attrib={"type": self.type.value, "condition": self.condition},
        )

    @classmethod
    def from_xml(cls, tag: ET.Element):
        cond_type = tag.get("type", ConditionType.IF.value)
        condition = tag.get("condition", "")

        result = SimulatorRuleCondition(type=ConditionType(cond_type))
        result.condition = condition

        return result
